#!/usr/bin/env python
import numpy
import numpy as np
import scipy.special

_alpha = 0.32


def interval(best,lo=np.nan,hi=np.nan):
    """
    Pythonized interval for easy output to yaml
    """
    return [float(best),[float(lo),float(hi)]]

def mean_interval(data, alpha=_alpha):
    """
    Interval assuming gaussian posterior.
    """
    mean =np.mean(data)
    sigma = np.std(data)
    scale = scipy.stats.norm.ppf(1-alpha/2.)
    return interval(mean,mean-scale*sigma,mean+scale*sigma)


def median_interval(data, alpha=_alpha):
    """
    Median including bayesian credible interval.
    """
    q = [100*alpha/2., 50, 100*(1-alpha/2.)]
    lo,med,hi = numpy.percentile(data,q)
    return interval(med,lo,hi)
    
def peak(data, bins=100):
    num,edges = np.histogram(data,bins=bins)
    centers = (edges[1:]+edges[:-1])/2.
    return centers[np.argmax(num)]

def kde_peak(data, samples=1000):
    """
    Identify peak using Gaussian kernel density estimator.
    """
    return kde(data,samples)[0]

def kde(data, samples=1000):
    """
    Identify peak using Gaussian kernel density estimator.
    """
    # Clipping of severe outliers to concentrate more KDE samples in the parameter range of interest
    mad = np.median(np.fabs(np.median(data) - data))
    cut = (data > np.median(data) - 5. * mad) & (data < np.median(data) + 5. * mad)
    x = data[cut]
    kde = scipy.stats.gaussian_kde(x)
    # No penalty for using a finer sampling for KDE evaluation except computation time
    values = np.linspace(np.min(x), np.max(x), samples) 
    kde_values = kde.evaluate(values)
    peak = values[np.argmax(kde_values)]
    return values[np.argmax(kde_values)], kde.evaluate(peak)


def peak_interval(data, alpha=_alpha, samples=1000):
    """
    Identify interval using Gaussian kernel density estimator.
    """
    peak = kde_peak(data,samples)
    x = np.sort(data.flat); n = len(x)
    # The number of entries in the interval
    window = int(np.rint((1.0-alpha)*n))
    # The start, stop, and width of all possible intervals
    starts = x[:n-window]; ends = x[window:]
    widths = ends - starts
    # Just the intervals containing the peak
    select = (peak >= starts) & (peak <= ends)
    widths = widths[select]
    if len(widths) == 0:
        raise ValueError('Too few elements for interval calculation')
    min_idx = np.argmin(widths)
    lo = x[min_idx]
    hi = x[min_idx+window]
    return interval(peak,lo,hi)

def min_interval(data, alpha=_alpha):
    x = np.sort(data.flat); n = len(x)
    # The number of entries in the interval
    window = int(np.rint((1.0-alpha)*n))
    # The start, stop, and width of all possible intervals
    starts = x[:n-window]; ends = x[window:]
    widths = ends - starts
    if len(widths) == 0:
        raise ValueError('Too few elements for interval calculation')
    min_idx = np.argmin(widths)
    lo = x[min_idx]
    hi = x[min_idx+window]
    mean = (hi+lo)/2.
    return interval(mean,lo,hi)


def norm_cdf(x):
    # Faster than scipy.stats.norm.cdf
    #https://en.wikipedia.org.wiki/Normal_distribution
    return 0.5*(1 + scipy.special.erf(x/np.sqrt(2)))

def random_pdf(value,pdf,size=None):
    if size is None: size = 1.0
    cdf = np.cumsum(pdf)
    cdf /= cdf[-1]
    fn = scipy.interpolate.interp1d(cdf, range(0, len(cdf)))
    index = np.rint(fn(np.random.uniform(size=size))).astype(int)
    return value[index]

def sky(lon=None,lat=None,size=1):
    """
    Outputs uniform points on sphere from:
    [0 < lon < 360] & [-90 < lat < 90]
    """
    if lon is None:
        umin,umax = 0,1
    else:
        lon = np.asarray(lon)
        lon = np.radians(lon + 360.*(lon<0))
        if   lon.size==1: umin=umax=lon/(2*np.pi)
        elif lon.size==2: umin,umax=lon/(2*np.pi)
        else: raise Exception('...')
        
    if lat is None:
        vmin,vmax = -1,1
    else:
        lat = np.asarray(lat)
        lat = np.radians(90 - lat)
        if   lat.size==1: vmin=vmax=np.cos(lat)
        elif lat.size==2: vmin,vmax=np.cos(lat)
        else: raise Exception('...')

    phi = 2*np.pi*np.random.uniform(umin,umax,size=size)
    theta = np.arcsin(np.random.uniform(vmin,vmax,size=size))
    return np.degrees(phi),np.degrees(theta)

if __name__ == "__main__":
    import argparse
    description = "python script"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('args',nargs=argparse.REMAINDER)
    opts = parser.parse_args(); args = opts.args

    import pylab as plt
    ax=plt.subplot(221,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3)
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(222,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lat=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(223,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lon=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)

    ax=plt.subplot(224,projection='aitoff')
    ax.grid(True)
    lon,lat = sky(size=1e3,lon=[0,45],lat=[30,45])
    lon,lat= np.radians([lon-360.*(lon>180),lat])
    ax.scatter(lon,lat,marker='.',s=2)



"""
Classes which manage object catalogs live here.
"""

import numpy
import numpy as np
import pyfits
import healpy
import copy

import ugali.utils.projector

from ugali.utils.config import Config
from ugali.utils.projector import gal2cel,cel2gal
from ugali.utils.healpix import ang2pix,superpixel
from ugali.utils.logger import logger
############################################################
### ADW: This needs to be rewritten to use fitsio
############################################################

class Catalog:

    def __init__(self, config, roi=None, data=None):
        """
        Class to store information about detected objects.

        The raw data from the fits file is stored. lon and lat are derived quantities
        based on chosen coordinate system.

        INPUTS:
            config: Config object
            roi[None] : Region of Interest to load catalog data for
            data[None]: pyfits table data (fitsrec) object.
        """
        #self = config.merge(config_merge) # Maybe you would want to update parameters??
        self.config = Config(config)

        if data is None:
            self._parse(roi)
        else:
            self.data = data

        self._defineVariables()

    def __add__(self, other):
        return mergeCatalogs([self,other])

    def __len__(self):
        return len(self.objid)

    def applyCut(self, cut):
        """
        Return a new catalog which is a subset of objects selected using the input cut array.
        """
        return Catalog(self.config, data=self.data[cut])

    def bootstrap(self, mc_bit=0x10, seed=None):
        """
        Return a random catalog by boostrapping the colors of the objects in the current catalog.
        """
        if seed is not None: numpy.random.seed(seed)
        data = copy.deepcopy(self.data)
        idx = numpy.random.randint(0,len(data),len(data))
        data[self.config['catalog']['mag_1_field']][:] = self.mag_1[idx]
        data[self.config['catalog']['mag_err_1_field']][:] = self.mag_err_1[idx]
        data[self.config['catalog']['mag_2_field']][:] = self.mag_2[idx]
        data[self.config['catalog']['mag_err_2_field']][:] = self.mag_err_2[idx]
        data[self.config['catalog']['mc_source_id_field']][:] |= mc_bit
        return Catalog(self.config, data=data)

    def project(self, projector = None):
        """
        Project coordinates on sphere to image plane using Projector class.
        """
        if projector is None:
            try:
                self.projector = ugali.utils.projector.Projector(self.config['coords']['reference'][0],
                                                                 self.config['coords']['reference'][1])
            except KeyError:
                logger.warning('Projection reference point is median (lon, lat) of catalog objects')
                self.projector = ugali.utils.projector.Projector(numpy.median(self.lon), numpy.median(self.lat))
        else:
            self.projector = projector

        self.x, self.y = self.projector.sphereToImage(self.lon, self.lat)

    def spatialBin(self, roi):
        """
        Calculate indices of ROI pixels corresponding to object locations.
        """
        if hasattr(self,'pixel_roi_index') and hasattr(self,'pixel'): 
            logger.warning('Catalog alread spatially binned')
            return

        # ADW: Not safe to set index = -1 (since it will access last entry); 
        # np.inf would be better...
        self.pixel = ang2pix(self.config['coords']['nside_pixel'],self.lon,self.lat)
        self.pixel_roi_index = roi.indexROI(self.lon,self.lat)

        if numpy.any(self.pixel_roi_index < 0):
            logger.warning("Objects found outside ROI")

    def write(self, outfile):
        """
        Write the current object catalog to fits file.
        """
            
        hdu = pyfits.BinTableHDU(self.data)
        hdu.writeto(outfile, clobber=True)

    def _parse(self, roi=None):
        """
        Helper function to parse a catalog file and return a pyfits table.

        CSV format not yet validated.

        !!! Careful, reading a large catalog is memory intensive !!!
        """
        
        filenames = self.config.getFilenames()

        if len(filenames['catalog'].compressed()) == 0:
            raise Exception("No catalog file found")
        elif roi is not None:
            pixels = roi.getCatalogPixels()
            self.data = readCatalogData(filenames['catalog'][pixels])
        elif len(filenames['catalog'].compressed()) == 1:
            file_type = filenames[0].split('.')[-1].strip().lower()
            if file_type == 'csv':
                self.data = numpy.recfromcsv(filenames[0], delimiter = ',')
            elif file_type in ['fit', 'fits']:
                self.data = pyfits.open(filenames[0])[1].data
            else:
                logger.warning('Unrecognized catalog file extension %s'%(file_type))
        else:
            self.data = readCatalogData(filenames['catalog'].compressed())


        # ADW: This is horrible and should never be done...
        selection = self.config['catalog'].get('selection')
        if not selection:
            pass
        elif 'self.data' not in selection:
            msg = "Selection does not contain 'data'"
            raise Exception(msg)
        else:
            logger.warning('Evaluating selection: \n"%s"'%selection)
            sel = eval(selection)
            self.data = self.data[sel]

        #print 'Found %i objects'%(len(self.data))

    def _defineVariables(self):
        """
        Helper funtion to define pertinent variables from catalog data.
        """
        self.objid = self.data.field(self.config['catalog']['objid_field'])
        self.lon = self.data.field(self.config['catalog']['lon_field'])
        self.lat = self.data.field(self.config['catalog']['lat_field'])

        #if self.config['catalog']['coordsys'].lower() == 'cel' \
        #   and self.config['coords']['coordsys'].lower() == 'gal':
        #    logger.info('Converting catalog objects from CELESTIAL to GALACTIC cboordinates')
        #    self.lon, self.lat = ugali.utils.projector.celToGal(self.lon, self.lat)
        #elif self.config['catalog']['coordsys'].lower() == 'gal' \
        #   and self.config['coords']['coordsys'].lower() == 'cel':
        #    logger.info('Converting catalog objects from GALACTIC to CELESTIAL coordinates')
        #    self.lon, self.lat = ugali.utils.projector.galToCel(self.lon, self.lat)

        self.mag_1 = self.data.field(self.config['catalog']['mag_1_field'])
        self.mag_err_1 = self.data.field(self.config['catalog']['mag_err_1_field'])
        self.mag_2 = self.data.field(self.config['catalog']['mag_2_field'])
        self.mag_err_2 = self.data.field(self.config['catalog']['mag_err_2_field'])

        if self.config['catalog']['mc_source_id_field'] is not None:
            if self.config['catalog']['mc_source_id_field'] in self.data.names:
                self.mc_source_id = self.data.field(self.config['catalog']['mc_source_id_field'])
                logger.info('Found %i MC source objects'%(numpy.sum(self.mc_source_id > 0)))
            else:
                #ADW: This is pretty kludgy, please fix... (FIXME)
                columns_array = [pyfits.Column(name = self.config['catalog']['mc_source_id_field'],
                                               format = 'I',
                                               array = numpy.zeros(len(self.data)))]
                hdu = pyfits.new_table(columns_array)
                self.data = pyfits.new_table(pyfits.new_table(self.data.view(np.recarray)).columns + hdu.columns).data
                self.mc_source_id = self.data.field(self.config['catalog']['mc_source_id_field'])

        # should be @property
        if self.config['catalog']['band_1_detection']:
            self.mag = self.mag_1
            self.mag_err = self.mag_err_1
        else:
            self.mag = self.mag_2
            self.mag_err = self.mag_err_2
            
        # should be @property
        self.color = self.mag_1 - self.mag_2
        self.color_err = numpy.sqrt(self.mag_err_1**2 + self.mag_err_2**2)

        logger.info('Catalog contains %i objects'%(len(self.data)))

    # This assumes Galactic coordinates
    @property
    def ra_dec(self): return gal2cel(self.lon,self.lat)
    @property
    def ra(self): return self.ra_dec[0]
    @property
    def dec(self): return self.ra_dec[1]

    @property
    def glon_glat(self): return self.lon,self.lat
    @property
    def glon(self): return self.lon
    @property
    def glat(self): return self.lat

############################################################

def mergeCatalogs(catalog_array):
    """
    Input is an array of Catalog objects. Output is a merged catalog object.
    Column names are derived from first Catalog in the input array.
    """
    len_array = []
    for ii in range(0, len(catalog_array)):
        len_array.append(len(catalog_array[ii].data))
    cumulative_len_array = numpy.cumsum(len_array)
    cumulative_len_array = numpy.insert(cumulative_len_array, 0, 0)

    columns = pyfits.new_table(catalog_array[0].data).columns
    hdu = pyfits.new_table(columns, nrows=cumulative_len_array[-1])
    for name in columns.names:
        for ii in range(0, len(catalog_array)):
            if name not in catalog_array[ii].data.names:
                continue
            hdu.data.field(name)[cumulative_len_array[ii]: cumulative_len_array[ii + 1]] = catalog_array[ii].data.field(name)

    catalog_merged = Catalog(catalog_array[0].config, data=hdu.data)
    return catalog_merged


############################################################

def precomputeCoordinates(infile, outfile):
    import numpy.lib.recfunctions as rec
    hdu = pyfits.open(infile)
    data = hdu[1].data
    columns = hdu[1].columns
    names = [n.lower() for n in hdu[1].data.names]

    if 'glon' not in names and 'glat' not in names:
        logger.info("Writing 'GLON' and 'GLAT' columns")
        glon, glat = ugali.utils.projector.celToGal(data['RA'], data['DEC'])
        out = rec.append_fields(data,['GLON','GLAT'],[glon,glat],
                                usemask=False,asrecarray=True)
    elif 'ra' not in names and 'dec' not in names:
        logger.info("Writing 'RA' and 'DEC' columns")
        ra, dec = ugali.utils.projector.galToCel(data['GLAT'], data['GLON'])
        out = rec.append_fields(data,['RA','DEC'],[ra,dec],
                                usemask=False,asrecarray=True)
    
    hdu_out = pyfits.BinTableHDU(out)
    hdu_out.writeto(outfile, clobber=True)

############################################################

def readCatalogData(infiles):
    """ Read a set of catalog FITS files into a single recarray. """
    if isinstance(infiles,basestring): infiles = [infiles]
    data, len_data = [],[]
    for f in infiles:
        data.append(pyfits.open(f)[1].data)
        len_data.append(len(data[-1]))

    cumulative_len_array = numpy.cumsum(len_data)
    cumulative_len_array = numpy.insert(cumulative_len_array, 0, 0)
    columns = data[0].columns
    table = pyfits.new_table(columns, nrows=cumulative_len_array[-1])

    for name in columns.names:
        for ii in range(0, len(data)):
            if name not in data[ii].names:
                raise Exception("Column %s not found in %"(name,infiles[ii]))
            table.data.field(name)[cumulative_len_array[ii]: cumulative_len_array[ii + 1]] = data[ii].field(name)
        
    return table.data

def makeHDU(config,mag_1,mag_err_1,mag_2,mag_err_2,lon,lat,mc_source_id):
    """
    Create a catalog fits file object based on input data.

    ADW: This should be combined with the write_membership function of loglike.
    """

    if config['catalog']['coordsys'].lower() == 'cel' \
       and config['coords']['coordsys'].lower() == 'gal':
        lon, lat = ugali.utils.projector.gal2cel(lon, lat)
    elif config['catalog']['coordsys'].lower() == 'gal' \
       and config['coords']['coordsys'].lower() == 'cel':
        lon, lat = ugali.utils.projector.cel2gal(lon, lat)

    columns = [
        pyfits.Column(name=config['catalog']['objid_field'],
                      format = 'D',array = np.arange(len(lon))),
        pyfits.Column(name=config['catalog']['lon_field'],
                      format = 'D',array = lon),
        pyfits.Column(name = config['catalog']['lat_field'],          
                      format = 'D',array = lat), 
        pyfits.Column(name = config['catalog']['mag_1_field'],        
                      format = 'E',array = mag_1),
        pyfits.Column(name = config['catalog']['mag_err_1_field'],    
                      format = 'E',array = mag_err_1),
        pyfits.Column(name = config['catalog']['mag_2_field'],        
                      format = 'E',array = mag_2),
        pyfits.Column(name = config['catalog']['mag_err_2_field'],    
                      format = 'E',array = mag_err_2),
        pyfits.Column(name = config['catalog']['mc_source_id_field'], 
                      format = 'I',array = mc_source_id),
    ]

    hdu = pyfits.new_table(columns)
    return hdu



#!/usr/bin/env python
"""
Spatial kernels for dwarf galaxies.
"""

import sys
import inspect
from abc import abstractmethod
from collections import OrderedDict as odict
import copy

import numpy
import numpy as np
import scipy.integrate
import scipy.interpolate
import healpy

import ugali.utils.projector
from ugali.utils.projector import Projector, angsep
from ugali.analysis.model import Model, Parameter
from ugali.utils.healpix import ang2vec, ang2pix, query_disc

from ugali.utils.logger import logger

#ADW: WARNING some funky stuff may be happening at the origin (0,0)

class Kernel(Model):
    """
    Base class for kernels.
    """
    _params = odict([
        ('lon',      Parameter(0.0, [0.0,  360.  ])),
        ('lat',      Parameter(0.0, [-90., 90.   ])),
    ])
    _mapping = odict([])
    _proj = 'ait'
    
    def __init__(self, proj='ait', **kwargs):
        # This __init__ is probably not necessary...
        self.proj = proj
        super(Kernel,self).__init__(**kwargs)

    def __call__(self, lon, lat):
        return self.pdf(lon, lat)

    @abstractmethod
    def _kernel(self, r):
        # unnormalized, untruncated kernel
        pass

    def _pdf(self, radius):
        # unnormalized, truncated kernel
        return np.where(radius<=self.edge, self._kernel(radius), 0.)

    @abstractmethod
    def pdf(self, lon, lat):
        # normalized, truncated pdf
        pass

    @property
    def norm(self):
        # Numerically integrate the pdf
        return 1./self.integrate()

    @property
    def projector(self):
        if self.proj is None or self.proj.lower()=='none':
            return None
        else:
            return Projector(self.lon, self.lat, self.proj)

    def integrate(self, rmin=0, rmax=numpy.inf):
        """
        Calculate the 2D integral of the 1D surface brightness profile 
        (i.e, the flux) between rmin and rmax (elliptical radii). 
        rmin : minimum integration radius (deg)
        rmax : maximum integration radius (deg)
        return : Solid angle integral (deg^2)
        """
        if rmin < 0: raise Exception('rmin must be >= 0')
        integrand = lambda r: self._pdf(r) * 2*numpy.pi * r
        return scipy.integrate.quad(integrand,rmin,rmax,full_output=True,epsabs=0)[0]

class ToyKernel(Kernel):
    """
    Simple toy kernel that selects healpix pixels within
    the given extension radius. Similar to 'RadialDisk'.
    """

    _params = odict(
        Kernel._params.items() + 
        [
            ('extension',     Parameter(0.1, [0.0001,5.0]) ), 
            ('nside',         Parameter(4096,[4096,4096])),
        ])

    def _cache(self, name=None):
        pixel_area = healpy.nside2pixarea(self.nside,degrees=True)
        vec = ang2vec(self.lon, self.lat)
        self.pix = query_disc(self.nside,vec,self.extension)
        self._norm = 1./(len(self.pix)*pixel_area)

    @property
    def norm(self):
        return self._norm

    def _pdf(self,pix):
        return  np.in1d(pix,self.pix)
        
    def pdf(self,lon,lat):
        pix = ang2pix(self.nside,lon,lat)
        return self.norm * self._pdf(pix)
       
class EllipticalKernel(Kernel):
    """
    Base class for elliptical kernels.
    Ellipticity is defined as 1 - b/a where a,b are the semi-major,semi-minor
    axes respectively. The position angle is defined in degrees east of north.
    This definition follows from Martin et al. 2008:
    http://adsabs.harvard.edu/abs/2008ApJ...684.1075M

    ### This is a depricated warning (2015/08/12)
    ### ADW: WARNING!!! This is actually the PA *WEST* of North!
    ### to get the conventional PA EAST of North take 90-PA
    ### Documentation?
    """
    _params = odict(
        Kernel._params.items() + 
        [
            ('extension',     Parameter(0.1, [0.0001,0.5]) ),   
            ('ellipticity',   Parameter(0.0, [0.0, 0.99]) ),    # Default 0 for RadialKernel
            ('position_angle',Parameter(0.0, [0.0, 180.0]) ),  # Default 0 for RadialKernel
            # This is the PA *WEST* of North.
            # to get the conventional PA EAST of North take 90-PA
            # Would it be better to have bounds [-90,90]?
        ])
    _mapping = odict([
        ('e','ellipticity'),
        ('theta','position_angle'),
    ])

    @property
    def norm(self):
        norm = super(EllipticalKernel,self).norm
        return  norm * 1./self.jacobian

    @property
    def jacobian(self):
        return 1. - self.e

    @property
    def a(self):
        return self.extension

    @property
    def b(self):
        return self.a*self.jacobian

    @property
    def edge(self):
        return 5.*self.extension

    def angsep(self,lon,lat):
        return angsep(self.lon,self.lat,lon,lat)
        
    def radius(self,lon,lat):
        x,y = self.projector.sphereToImage(lon,lat)
        costh = np.cos(np.radians(self.theta))
        sinth = np.sin(np.radians(self.theta))
        return np.sqrt(((x*costh-y*sinth)/(1-self.e))**2 + (x*sinth+y*costh)**2)
        
    def pdf(self,lon,lat):
        radius = self.radius(lon,lat)
        return self.norm*self._pdf(radius)

    def sample_radius(self, n):
        """
        Sample the radial distribution (deg) from the 2D stellar density.
        Output is elliptical radius in true projected coordinates.
        """
        edge = self.edge if self.edge<20*self.extension else 20*self.extension
        radius = np.linspace(0, edge, 1.e5)
        pdf = self._pdf(radius) * np.sin(np.radians(radius))
        cdf = np.cumsum(pdf)
        cdf /= cdf[-1]
        fn = scipy.interpolate.interp1d(cdf, range(0, len(cdf)))
        index = numpy.floor(fn(numpy.random.uniform(size=n))).astype(int)
        return radius[index]
 
    def sample_lonlat(self, n):
        """
        Sample 2D distribution of points in lon, lat
        """
        # From http://en.wikipedia.org/wiki/Ellipse#General_parametric_form
        # However, Martin et al. (2009) use PA theta "from North to East"
        # Definition of phi (position angle) is offset by pi/4
        # Definition of t (eccentric anamoly) remains the same (x,y-frame usual)
        # In the end, everything is trouble because we use glon, glat...

        radius = self.sample_radius(n)
        a = radius; b = self.jacobian * radius

        t = 2. * np.pi * numpy.random.rand(n)
        cost,sint = np.cos(t),np.sin(t)
        phi = np.pi/2. - np.deg2rad(self.theta)
        cosphi,sinphi = np.cos(phi),np.sin(phi)
        x = a*cost*cosphi - b*sint*sinphi
        y = a*cost*sinphi + b*sint*cosphi
        
        if self.projector  is None:
            logger.debug("Creating AITOFF projector for sampling")
            projector = Projector(self.lon,self.lat,'ait')
        else:
            projector = self.projector
        lon, lat = projector.imageToSphere(x, y)
        return lon, lat
 
    simulate = sample_lonlat
    sample = sample_lonlat

    # Back-compatibility
    def setExtension(self,extension):
        self.extension = extension

    def setCenter(self,lon,lat):
        self.lon = lon
        self.lat = lat

class EllipticalDisk(EllipticalKernel):
    """
    Simple uniform disk kernel for testing.
    f(r) = 1  for r <= r_0
    f(r) = 0  for r > r_0
    """
    _params = EllipticalKernel._params
    _mapping = odict(
        EllipticalKernel._mapping.items() +
        [
            ('r_0','extension')
        ])
                     
    ### ADW: stellar mass conversion?
    def _kernel(self, radius):
        return np.where(radius<=self.r_0, 1.0, 0.0) 
 
    @property
    def norm(self):
        return 1./(np.pi*self.r_0**2 * self.jacobian)
 
class EllipticalGaussian(EllipticalKernel):
    """
    Simple Gaussian kernel for testing:
    f(r) = C * exp(-r**2 / 2*sigma**2)
    """
    _params = EllipticalKernel._params
    _mapping = odict(
        EllipticalKernel._mapping.items() +
        [
            ('sigma','extension')
        ])
 
    ### ADW: stellar mass conversion?
    def _kernel(self, radius):
        return np.exp(-radius**2/(2*self.sigma**2))
 
    @property
    def norm(self):
        # Analytic integral from 0 to edge
        return 1./(2*np.pi*self.sigma**2*(1-np.exp(-self.edge**2/(2*self.sigma**2)))*self.jacobian)
 
class EllipticalExponential(EllipticalKernel):
    """
    Stellar density distribution for Exponential profile:
    f(r) = C * exp(-r / r_e)
    where r_e = r_h/1.68
    http://adsabs.harvard.edu//abs/2006MNRAS.365.1263M (Eq. 5)        
    """
    _params = odict(EllipticalKernel._params)
    _mapping = odict(
        EllipticalKernel._mapping.items() +
        [
            ('r_h','extension'), # Half-light radius
        ])
    
    def _kernel(self,radius):
        return np.exp(-radius/self.r_e)
 
    @property
    def norm(self):
        # Analytic integral
        xedge = self.edge/self.r_e
        return 1./(2*np.pi*self.r_e**2*(1 - (xedge+1)*np.exp(-xedge))*self.jacobian)
 
    @property
    def r_e(self):
        # Exponential scale radius
        return self.r_h/1.68
 
    @property
    def edge(self):
        return 20.*self.r_h
 
class EllipticalPlummer(EllipticalKernel):
    """
    Stellar density distribution for Plummer profile:
    f(r) = C * r_c**2 / (r_c**2 + r**2)**2
    http://adsabs.harvard.edu//abs/2006MNRAS.365.1263M (Eq. 6)
    """
    _params = odict(
        EllipticalKernel._params.items() + 
        [
            ('truncate', Parameter(3.0, [0.0, np.inf]) ), # Truncation radius
        ])
    _mapping = odict(
        EllipticalKernel._mapping.items() +
        [
            ('r_c','extension'), # Plummer radius
            ('r_h','extension'), # ADW: Depricated
            ('r_t','truncate'),  # Tidal radius
        ])
 
    def _kernel(self, radius):
        return 1./(numpy.pi*self.r_h**2 * (1.+(radius/self.r_h)**2)**2)

    def _cache(self, name=None):
        if name in [None,'extension','ellipticity','truncate']:
            self._norm = 1./self.integrate() * 1./self.jacobian
        else:
            return

    @property
    def norm(self):
        return self._norm

    @property
    def u_t(self):
        # Truncation factor
        return self.truncate/self.extension
 
    @property
    def edge(self):
        return self.r_t
 
class EllipticalKing(EllipticalKernel):
    """
    Stellar density distribution for King profile:
    f(r) = C * [ 1/sqrt(1 + (r/r_c)**2) - 1/sqrt(1 + (r_t/r_c)**2) ]**2
    http://adsabs.harvard.edu//abs/2006MNRAS.365.1263M (Eq. 4)
    
    The half-light radius is related to the King radius for 
    c = log10(r_t/r_c) = 0.7 :
    r_h = 1.185 * r_c
    http://adsabs.harvard.edu/abs/2010MNRAS.406.1220W (App.B)
    """
    _params = odict(
        EllipticalKernel._params.items() + 
        [
            ('truncate', Parameter(3.0, [0.0, np.inf]) ), # Truncation radius
        ])
    _mapping = odict(
        EllipticalKernel._mapping.items() +
        [
            ('r_c','extension'), # Core radius
            ('r_t','truncate'),  # Tidal radius
        ])
 
    def _kernel(self, radius):
        return ((1./np.sqrt(1.+(radius/self.r_c)**2))-(1./np.sqrt(1.+(self.r_t/self.r_c)**2)))**2

    def _cache(self, name=None):
        if name in ['extension','ellipticity','truncate']:
            self._norm = 1./self.integrate()
        else:
            return

    @property
    def norm(self):
        return self._norm
 
    @property
    def c(self):
        return np.log10(self.r_t/self.r_c)
 
    @property
    def edge(self):
        return self.r_t

#####################################################
### Radial Kernels
#####################################################
 
class RadialKernel(EllipticalKernel):
    """
    Radial kernel subclass fixing ellipticity and 
    position angle to zero.
    """
    _fixed_params = ['ellipticity','position_angle']

    def __init__(self,**kwargs):
        # This is a bit messy because the defaults are set
        # at the instance level not at the class level
        self._params = copy.deepcopy(self._params)
        self._params['ellipticity'].set(0, [0, 0])
        self._params['position_angle'].set(0, [0, 0])
        #logger.warning("Setting bounds on extension")
        #self._params['extension'].set(0.1, [1e-4, 0.1])

        super(RadialKernel,self).__init__(**kwargs)
        
    def pdf(self, lon, lat):
        if self.projector is None:
            radius = angsep(self.lon,self.lat,lon,lat)
            return self.norm*self._pdf(radius)
        else:
            return super(RadialKernel,self).pdf(lon,lat)

    # Back-compatibility
    def surfaceIntensity(self,radius):
        return self.norm*self._pdf(radius)
        
# For fast access...
class RadialDisk(RadialKernel,EllipticalDisk): pass
class RadialExponential(RadialKernel,EllipticalExponential): pass
class RadialGaussian(RadialKernel,EllipticalGaussian): pass
class RadialPlummer(RadialKernel,EllipticalPlummer): pass
class RadialKing(RadialKernel,EllipticalKing): pass
 
Disk        = RadialDisk
Gaussian    = RadialGaussian
Exponential = RadialExponential
Plummer     = RadialPlummer
King        = RadialKing

### def kernelFactory2(name, **kwargs):
###     """
###     Factory for creating spatial kernels. Arguments
###     are passed directly to the constructor of the chosen
###     kernel.
###     """
###     fn = lambda member: inspect.isclass(member) and member.__module__==__name__
###     kernels = odict(inspect.getmembers(sys.modules[__name__], fn))
###  
###     if name not in kernels.keys():
###         msg = "%s not found in kernels:\n %s"%(name,kernels.keys())
###         logger.error(msg)
###         msg = "Unrecognized kernel: %s"%name
###         raise Exception(msg)
###  
###     return kernels[name](**kwargs)
###  
###  
### def kernelFactory(name, **kwargs):
###     """
###     Factory for creating spatial kernels. Arguments
###     are passed directly to the constructor of the chosen
###     kernel.
###     """
###     fn = lambda member: inspect.isclass(member) and member.__module__==__name__
###     kernels = odict(inspect.getmembers(sys.modules[__name__], fn))
###  
###     if name not in kernels.keys():
###         msg = "%s not found in kernels:\n %s"%(name,kernels.keys())
###         logger.error(msg)
###         msg = "Unrecognized kernel: %s"%name
###         raise Exception(msg)
###  
###     return kernels[name](**kwargs)

#ADW: Should change 'name' to 'type' (but lots of legacy to deal with)
def factory(name, **kwargs):
    from ugali.utils.factory import factory
    return factory(name, module=__name__, **kwargs)

kernelFactory = factory


if __name__ == "__main__":
    import argparse
    description = "python script"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('args',nargs=argparse.REMAINDER)
    opts = parser.parse_args(); args = opts.args
